/* gencode.c
	code generator som-c		2 Jan 2004
	som3   						4 Nov 2004
	som-v16						7 Nov 2005
	som-v17						24 Dec 2005
*/

#include "compile.h"

#define CSSIZE	(MAXCS - MAXDS)	// CS[0..CSSIZE-1]
#define MACSIZE 5000			// space for macro
#define NUMBER	300				// denote arg as number

extern int cmode;				// compile mode

int CP = 3;						// current code pointer
int currentf;					// current function idx

int vmin = 0, vmax = 0;			// number of locals
int v300;						// keep constant value
ventry vrec[LVMAX];				// v allocation record

// CPX is used to store macro and imm code
// it is "after" the normal code in CS[]
PRIVATE int CPX = CSSIZE-MACSIZE;

#define item2(x) 	car(cdr(x))
#define item3(x)	car(cdr(cdr(x)))
#define is8(n)		(n >= 0 && n < 256)
#define is16(n) 	(n >= -32767 && n < 32768)

PRIVATE int atomeq(int e, int type, int value){
	return (car(e) == type) && (cdr(e) == value);
}

// free only the temp
void freev(int v){
	if(v > vmin && v < LVMAX) vrec[v].free = 0;
}

// allocate a new temp
PRIVATE int newv(void){
	int i;
	for(i=vmin+1; i<=vmax; i++)
		if(vrec[i].free == 0 ){
			vrec[i].free = 1;
			return i;
		}
	vmax++;
	vrec[vmax].pair = vmax;	// no pair
	vrec[vmax].free = 1;
	return vmax;
}

// genex, if number out movi/lit
int genex2(int e){
	int v;
	v = genex(e);
	if( v == NUMBER ){
		if( is16(v300) ){
			v = newv();
			outa2(tMovi,v,v300);
		}else{
			outa(tLit,v300);
			v = RETVAL;
		}
	}
	return v;
}

PRIVATE int genbop(int op, int e){
	int v, v1, v2;
	v1 = genex2(car(e));
	v2 = genex(item2(e));
	v = newv();
	if( v2 == NUMBER){
		if( is8(v300) ){
			outa3(op+16,v,v1,v300);	// imm mode
		}else if( is16(v300)){
			outa2(tMovi,v,v300);
			outa3(op,v,v1,v);
		}else{
			outa(tLit,v300);
			outa3(op,v,v1,0);
		}
	}else
		outa3(op,v,v1,v2);
	freev(v1);
	freev(v2);
	return v;
}

// gen num to v using mov/pass, movi/passi
PRIVATE void gennum(int op, int v, int n){
	if( is16(n) ){
		outa2(op+1,v,n);		// movi/passi
	}else{
		outa(tLit,n);
		outa3(op,v,RETVAL,0);	// mov/pass
	}
}

// if there is any break, patch jump to CP
PRIVATE void patchbreak(void){
	int ads;
	if( ytos() == MARK ){	// TO without break
		ypop();
		return;
	}
	while( ytos() == BMARK ){
		ypop();			// throw away BMARK
		ads = ypop();
		patch(ads,CP);
	}
	if( ytos() != MARK )
		seterror("patch-break no matching MARK");
	ypop();				// throw away MARK
}

// genex e and put the result in v
PRIVATE void genput(int v, int e){
	int v1, op, a;
	v1 = genex(e);
	if(v1 == NUMBER){
		gennum(tMov,v,v300);
	}else if( isatom(e) ){
		outa3(tMov,v,v1,0);
	}else{						// v2 has op
		decode(CP-1,&op,&a);	// special case for "call","sys"
		if( op == tCall || op == tCallt || op == tSys)
			outa3(tMov,v,RETVAL,0);
		else
			patcha1(CP-1,v);
	}
	freev(v1);
}

// gen e, put in v
// not optimise, some extra mov, to be improved
PRIVATE void genput2(int v, int e){
	int v1;
	v1 = genex(e);
	if(v1 == NUMBER)
		gennum(tMov,v,v300);
	else
		outa3(tMov,v,v1,0);
	freev(v1);
}

// e = (block ex ... ex), ex = (num ex)
// get lo and hi labels from caselist
PRIVATE void getcaselist(int e, int *lo, int *hi){
	int n, ex;
	*lo = 0;
	*hi = 0;
	if( e == NIL ) return;
	e = cdr(e);
	ex = car(e);  		// a = (num ex)
	n = cdr(car(ex));  	// n = num.ref
	*lo = n;
	*hi = n;
	while( e != NIL ){
		ex = car(e);
		n = cdr(car(ex));
		if( n < 0 ) break;  // else case
		if( n < *lo ) *lo = n;
		if( n > *hi ) *hi = n;
		e = cdr(e);
	}
}

// high quality, e = (ex0 block)
PRIVATE int gencase(int e){
	int lo, hi, ads, ads2, i, c, a;
	int op, d, ref;
	int v, v1, e1;

	v1 = genex2(car(e));		// gen ex0
	ads = CP;
	outa2(tCase,v1,0);
	freev(v1);
	e = item2(e);				// block
	getcaselist(e,&lo,&hi);
	outa(tLit,lo);
	outa(tLit,hi);
	for(i=lo; i<=hi; i++)
		outa(tJmp,0);			// empty jump table
	c = 0;						// c is backchained
	e = cdr(e);					// e = (ex .. ex)
	v = newv();
	while( e != NIL ){
		a = car(e);				// a = (num ex)
		ref = cdr(car(a)); 		// ref = num label
		if( ref > 0 ){			// skip else case ref-1
			ads2 = ref-lo+ads+3;// entry in jump table
			patch(ads2,CP);		// to jump here
			genput(v,item2(a));
			outa(tJmp,c);		// jump to eloc, backchain
			c = CP-1;
		}
		e = cdr(e);
	}
	// here is exelse
	// patch the empty entry in jumptable to exelse
	for(i=ads+3; i<=ads+3+hi-lo; i++){
		decode(i,&op,&d);
		if(d == 0) patch(i,CP);
	}
	patchd(ads,CP-ads);	// else case
	if( ref == -1 )		// gen else case
		genput(v,item2(a));
	// here is eloc
	// update backchain jmp to eloc
	while( c != 0 ){	// update jmp to eloc
		decode(c,&op,&d);
		patch(c,CP);
		c = d;
	}
	return v;
}

// e = body of fun (block ex ex ...)
PRIVATE void genfun(int idx, int e){
	int i, v, arity;
	setRef(idx,CP);
	vmin = getLv(idx);
	arity = getArg(idx);
	vmax = vmin;
	for(i=1; i<=vmax; i++){
		vrec[i].pair = i;
		vrec[i].free = 1;
	}
	outa3(tFun,arity,vmax,0);	// clean will update it
	ypush(MARK);
	v = genex2(e);
	patchbreak();
	outa3(tRet,v,vmax,0);
	clean(idx);
}

// update symbol table for static array and string
PRIVATE void updatesym(int ref, int ex){
	if(isatom(ex)){				// ex = string or pointer
		if(car(ex) == STRING){
			setArg(ref,VSTRING);
		}else if(car(ex) == ADS){
			setArg(ref,VARRAY);
		}
	}else if(isatom(car(ex))){	// ex = (array ..)
		if(atomeq(car(ex),OPER,tkARRAY)){
			setArg(ref,VARRAY);
		}
	}
}

// vector e = (name ex)
PRIVATE int genvec(int e, int *idx){
	int nm, v;
	*idx = genex2(item2(e));// vec index
	nm = car(e);
	v = cdr(nm);			// ref
	if( car(nm) == GNAME )	// car nm is type
		v = getRef(v);
	return v;
}

PRIVATE int genatom(int x){
	int type, ref, v;
	type = car(x);
	ref = cdr(x);
	v = 0;
	switch(type){
	case STRING:
	case ADS:
		outa(tAds,ref);
		v = RETVAL;
		break;
	case NUM:
		v = NUMBER;
		v300 = ref;
		break;
	case GNAME: v = getRef(ref); break;
	case LNAME: v = ref; break;
	case OPER:
		if(ref == tkBREAK){		// gen break
			ypush(CP);
			ypush(BMARK);
			outa(tJmp,0);
		}
		break;
	}
	return v;
}

int genex(int x){
	int a, e, e1, ads, ads2, v1, v2, v;
	int idx, nm, type, ref;

	if( x == NIL ) return 0;
	if(isatom(x)) return genatom(x);
	// x is list
	a = car(x);			// (op ex .. )
	e = cdr(x);			// e = (ex .. )
	if(car(a) != OPER)
		seterror("genex: expect operator");
	switch(cdr(a)){
	case tkPLUS: return genbop(tAdd,e);
	case tkMINUS: return genbop(tSub,e);
	case tkSTAR: return genbop(tMul,e);
	case tkSLASH: return genbop(tDiv,e);
	case tkAND: return genbop(tAnd,e);
	case tkBAR: return  genbop(tOr,e);
	case tkCARET: return genbop(tXor,e);
	case tkEQEQ: return genbop(tEq,e);
	case tkNE: return genbop(tNe,e);
	case tkLT: return genbop(tLt,e);
	case tkLE: return genbop(tLe,e);
	case tkGE: return genbop(tGe,e);
	case tkGT: return genbop(tGt,e);
	case tkMOD: return genbop(tMod,e);
	case tkGTGT: return genbop(tShr,e);
	case tkLTLT: return genbop(tShl,e);
	case tkNOT:
		v1 = genex2(car(e));
		freev(v1);
		v = newv();
		outa3(tNot,v,v1,0);
		return v;
	case tkARRAY:
		v1 = genex2(car(e));
		freev(v1);
		outa3(tSys,v1,0,SYSARRAY);
		return RETVAL;
	case tkSYSCALL: 			// e = (num ex0 ex0)
		a = car(e);				// sysnum
		if( car(a) != NUM )
			seterror("syscall expect number");
		e = cdr(e);				// at most two args
		v1 = genex2(car(e));
		v2 = genex2(item2(e));
		outa3(tSys,v1,v2,cdr(a)); // cdr a get num
		freev(v1);
		freev(v2);
		return RETVAL;
	case tkBB: 					// (block ex .. ex)
		v = 0;
		while( e != NIL ){		// e = (ex .. ex)
			freev(v);			// keep only the last one
			v = genex2(car(e));
			e = cdr(e);
		}
		return v;
	case tkBE:					// (call name ex ... ex)
		idx = cdr(car(e));		// e = (name ex .. ex)
		e = cdr(e);
		v1 = 1;
		while( e != NIL ){
			v = genex(car(e));
			if(v == NUMBER)
				gennum(tPass,v1,v300);
			else{
				outa3(tPass,v1,v,0);
				freev(v);
			}
			v1++;
			e = cdr(e);
		}
		a = getRef(idx);
		if( a == 0 )
			outa(tCallf,idx);
		else
			outa(tCall,a);
		return RETVAL;
	case tkRBRACKET:			// (mac name ex ... ex)
		idx = cdr(car(e));		// e = (name ex ... ex)
		v = domacro(idx,cdr(e));
		return v;
	case tkTO:					// (to name ex)
		idx = cdr(car(e));		// e = (name ex)
		currentf = idx;
		e = item2(e);
		if( e != NIL )
			genfun(idx,e);
		return 0;
	case tkCOLON:				// def macro
		idx = cdr(car(e));		// e = (name ex)
		currentf = idx;
		e = item2(e);
		ads = CP;
		CP = CPX;
		setRef(idx,CP);
		if( e != NIL ){
			a = getLv(idx) - getArg(idx); // local
			if( a == 0 ){
				setType(idx,tyMAC);
				setRef(idx,e);	// to parse tree
			}else
				setType(idx,tyFULL);
			// macro must not be recursive!
			ypush(MARK);
			genex2(e);
			patcha2(ads,vmax);	// update lv
			patchbreak();
			outa(tEnd,0);		// end marker
		}
		CPX = CP;
		CP = ads;
		return 0;
	case tkEQ:				// (= var/vec ex)
		a = car(e);			// e = (var/vec ex)
		e1 = item2(e);		// ex  RHS
		if(isatom(a)){		// var
			type = car(a);
			v = cdr(a);
			if(type == GNAME){
				if(cmode == 1)
					updatesym(v,e1);
				v = getRef(v);
			}
			genput(v,e1);
		}else{				// vec
			// cdr(a) is (name ex), vec LHS
			v = genvec(cdr(a),&v1);	// v1 = idx
			v2 = genex2(e1);		// RHS
			outa3(tStx,v2,v,v1);
			freev(v1);
			freev(v2);
		}
		return v;
	case tkLBRACKET: 		// vec RHS
		v2 = genvec(e,&v1);
		v = newv();
		outa3(tLdx,v,v2,v1);
		freev(v1);
		return v;
	case tkIF:				// (if e1 e)
		v = genex2(car(e));
		ads = CP;
		outa2(tJf,v,0);
		freev(v);
		v = genex2(item2(e));
		patchd(ads,CP-ads);
		return v;
	case tkELSE: 			// (ifelse e1 et ef)
		v1 = genex2(car(e));// e = (e1 et ef)
		ads = CP;
		outa2(tJf,v1,0);	// jf else
		freev(v1);
		v = newv();
		genput2(v,item2(e));// et
		outa(tJmp,0);
		patchd(ads,CP-ads);	// else
		ads = CP-1;
		genput2(v,item3(e)); // ef
		patch(ads,CP);
		return v;
	case tkWHILE: 			// (while e1 e)
		ypush(MARK);
		ads = CP;			// e = (e1 e)
		outa(tJmp,0);		// branch to cond
		v = genex2(item2(e));	// body of while
		ads2 = CP;
		v1 = genex2(car(e));	// cond
		outa2(tJt,v1,ads+1-CP); // jump to body
		patch(ads,ads2);
		patchbreak();
		freev(v1);
		return v;
	case tkFOR:				// (for lv ex0 ex0 ex)
		ypush(MARK);		// e = (lv ex0 ex0 ex)
		v1 = cdr(car(e));
		v2 = vrec[v1].pair;	// find adj(v1)
		if(v2 == v1){		// no pair
			v2 = newv();
			vrec[v1].pair = v2;
			vrec[v2].pair = 0;	// is a pair
		}
		e = cdr(e);			// e = (ex0 ex0 ex)
		v = genex(car(e));	// initial
		if(v == NUMBER)
			gennum(tMov,v1,v300-1);
		else
			outa3(tSubi,v1,v,1); // dec 1
		freev(v);
		genput(v2,item2(e)); // end
		ads = CP;
		outa(tJmp,0);		// jump to efor
		v = genex2(item3(e));// body of for
		patch(ads,CP);		// jmp test
		outa2(tEfor,v1,ads+1-CP);
		patchbreak();
		return v;
	case tkCASE:			// e = (ex0 block)
		return gencase(e);
	}
	return v;
}

// execute immediate line, use CS[CPX] area
void runimm(int e){
	int savecp, ads, v;
	if( e == NIL ) return;
	savecp = CP;
	CP = CPX;
	ads = CP;
	while( e != NIL ){		// e = (ex .. ex)
		v = genex2(car(e));
		freev(v);
		e = cdr(e);
	}
	outa(tEnd,0);
//	showCode2(ads);
	CPX = CP;
	CP = savecp;
	eval(ads);
}

// clean up callf, gen call to "main" if it is defined
// e is block of imm lines, execute it
void genfinal(void){
	int idx, d;
	patchCallf(1,CP-1);
	d = CP;
	CP = 1;
	idx = search("main");
	if(getType(idx) != tyNEW)
		outa(tCall,getRef(idx));
	outa(tEnd,0);
	CP = d;			// restore CP
}


