/* icode.c
	P. Chongstitvatana
	since R1						27 Aug 97
	start project RZ				3 Apr 2001
	project A1						22 Oct 2002
	project Som						13 Sept 2003
	code generator som-c			2 Jan 2004
	tail call elimination			Feb 2005
	som-v16 reLoc					7 Nov 2005
	som-v17							24 Dec 2005
*/

#include "compile.h"

void decode(int ip, int *code, int *data){
	*code = CS[ip] & 255;
	*data = CS[ip] >> 8;
}

void unpack(int a, int *a1, int *a2, int *a3){
	*a1 = a & 255;
	*a2 = (a >> 8) & 255;
	*a3 = (a >> 16) & 255;
}

void putcode(int ip, int code, int data){
	CS[ip] = (data << 8) | (code & 255);
}
/* format type of an op

A  a:24 op:8			// no rename
B  0 a2:8 a1:8 op:8		// rename a1 a2
C  a3:8 a2:8 a1:8 op:8	// rename a1 a2 a3
D  n:8 a2:8 a1:8 op:8 	// rename a1 a2
E  d:16 a1:8 op:8		// rename a1
F  0 a2:8 a1:8 op:8		// rename a2 only

A  call callt lit ads jmp fun
B  mov movd
C  bop ldx stx
D  bopi sys
E  jt jf efor movi case ret
F  pass

special case
  pass rename a2
  passi  no rename
  fun no rename
  ret rename a1

*/
PRIVATE char opfmt[] = {
	fC,fC,fC,fC,fC,fC,fC,fC,
	fC,fC,fC,fC,fC,fC,fC,fC,
	fD,fD,fD,fD,fD,fD,fD,fD,
	fD,fD,fD,fD,fD,fD,fD,fD,
	fB,fB,fE,fB,fF,fA,fA,fA,
	fC,fC,fA,fA,fA,fE,fE,fE,
	fA,fE,fE,fD,fA,fA
};
// print string of an op (for listing)
PRIVATE char opstr[][8] = {
	"Add","Sub","Mul","Div","Mod","And","Or","Xor",
	"Eq","Ne","Lt","Le","Gt","Ge","Shl","Shr",
	"Addi","Subi","Muli","Divi","Modi","Andi","Ori","Xori",
	"Eqi","Nei","Lti","Lei","Gti","Gei","Shli","Shri",
	"Not","Mov","Movi","Movd","Pass","Passi","Lit","Ads",
	"Ldx","Stx","Call","Callt","Fun","Ret","Efor","Case",
	"Jmp","Jt","Jf","Sys","End","Callf"
};

// get name if null return string "ref"
PRIVATE char *pN(int type, int ref){
	static char buf[40];
	char *nm;
	nm = findName(type,ref);
	if( nm != NULL ) return nm;
	sprintf(buf,"%d",ref);
	return buf;
}

void prCode(FILE *fo, int ip){
	int op, a, a1, a2, a3, d;
	decode(ip,&op,&a);
	if( op < EOC)
		fprintf(fo,"%s",opstr[op]);
	else
		fprintf(fo,"Undef");
	unpack(a,&a1,&a2,&a3);
	d = a >> 8;
	if(opfmt[op] == fD)					// imm 8-bit
		fprintf(fo," %d %d #%d",a1,a2,a3);
	else

	switch(op){
	case tFun:
		fprintf(fo," %d %d %s",a1,a2,pN(tyFUNC,ip));
		break;
	case tRet:
	case tNot:
	case tPass:
	case tMov: fprintf(fo," %d %d",a1,a2); break;
	case tPassi:
	case tMovi: fprintf(fo," %d #%d",a1,d); break;
	case tCall:
	case tCallt: fprintf(fo," %s",pN(tyFUNC,a)); break;
	case tCallf: fprintf(fo," %s",getName(a)); break;
	case tLit:
	case tAds:
	case tJmp: fprintf(fo," %d",a); break;
	case tEfor:
	case tJt:
	case tJf:
	case tCase:	fprintf(fo," %d %d",a1,ip+d); break;
	case tEnd: break;
	default: fprintf(fo," %d %d %d",a1,a2,a3);
	}
}

// output object code to file
void outCode(FILE *fo){
	int i, op, a, a1, a2, a3;
	fprintf(fo,"%d %d\n",1,CP-1);
	for(i=1; i<CP; i++){
		fprintf(fo,"%X ",CS[i]);
		if( (i % 5 ) == 0 ) fprintf(fo,"\n");
	}
}

void readCode(FILE *fi){
	int i, start, end, c;
	fscanf(fi,"%d%d",&start,&end);
	for(i=start; i<=end; i++){
		fscanf(fi,"%X",&c);
		CS[i] = c;
	}
}

void outData(FILE *fo){
	int i;
	fprintf(fo,"\n%d %d\n",MAXSYS,DS-1);
	for(i=MAXSYS; i<DS; i++){
		fprintf(fo,"%d ",M[i]);
		if( (i % 10) == 0) fprintf(fo,"\n");
	}
}

void readData(FILE *fi){
	int i, start, end, c;
	fscanf(fi,"%d%d",&start,&end);
	for(i=start; i<=end; i++){
		fscanf(fi,"%d",&c);
		M[i] = c;
	}
	DS = end+1;
}
// output listing to file
void showCode(FILE *fo){
	int i, op, a;
	for(i=1; i<CP; i++){
		fprintf(fo,"\n  %5d ",i);
		prCode(fo,i);
	}
	fprintf(fo,"\n");
}

// output listing for macro
void showCode2(int ip){
	int op, a;
	decode(ip,&op,&a);
	while( op != tEnd ){
		printf("\n  %5d ",ip);
		prCode(stdout,ip);
		ip++;
		decode(ip,&op,&a);
	}
	printf("\n");
}

/* out code with argument */
void outa( int code, int a1){
	CS[CP] = (a1 << 8 ) | (code & 255);
	CP++;
}

void outa2( int code, int a1, int a2){
	a1 = (a2 << 8) | (a1 & 255);
	CS[CP] = (a1 << 8 ) | (code & 255);
	CP++;
}

void outa3( int code, int a1, int a2, int a3){
	a1 = (a3 << 16) | ((a2 & 255) << 8) | (a1 & 255);
	CS[CP] = (a1 << 8 ) | (code & 255);
	CP++;
}

// change arg, preserve code
void patch( int ads, int v ){
	CS[ads] = (v << 8) | (CS[ads] & 255);
}
// patch d (disp 16 bits)
void patchd( int ads, int v ){
	CS[ads] = (v << 16) | (CS[ads] & 0x0ffff);
}
// patch arg a1
void patcha1( int ads, int v ){
	CS[ads] = ((v & 255) << 8) | (CS[ads] & 0x0ffff00ff);
}
// patch arg a2
void patcha2( int ads, int v ){
	CS[ads] = ((v & 255) << 16) | (CS[ads] & 0x0ff00ffff);
}

// rename if v is local
PRIVATE int rename2(int v){
	return (v > 0 && v < LVMAX) ? vrec[v].pair : v;
}

// base = 0 for currentf, base = vmax for macro expand
void rename3(int op, int a, int base, int *n){
	int a1, a2, a3;

	unpack(a,&a1,&a2,&a3);
	switch(opfmt[op]) {
//	case fA :			// a:24 op:8
//		break;
	case fB : 			// 0 a2:8 a1:8 op:8
	case fD :			// n:8 a2:8 a1:8 op:8
		a1 = rename2(a1+base);
		a2 = rename2(a2+base);
		break;
	case fC :			// a3:8 a2:8 a1:8 op:8
		a1 = rename2(a1+base);
		a2 = rename2(a2+base);
		a3 = rename2(a3+base);
		break;
	case fE : 			// d:16 a1:8 op:8
		a1 = rename2(a1+base);
		break;
	case fF :			// pass, rename a2
		a2 = rename2(a2+base);
		break;
	}
	*n = (a3<<16) | (a2<<8) | a1;
}

// the end of body is at CP
PRIVATE void reName(int fn){
	int i, j, a, c, ref, b;
	char lv2[LVMAX];

	for(i=1, j=1; i<=vmax; i++){	// shuffle pair
		a = vrec[i].pair;
		if(a == i){			// no pair
			lv2[j] = i;
			j++;
		}else if(a != 0){	// has pair
			lv2[j] = i;
			lv2[j+1] = a;	// record both
			j += 2;
		}
	}
	for(i=1; i<=vmax; i++)	// rename vrec[]
//		vrec[lv2[i]].pair = vmax - i + 1;
		vrec[lv2[i]].pair = i;

	ref = getRef(fn);
	for( i=ref+1; i<CP; i++ ) {
		decode(i,&c,&a);
		rename3(c,a,0,&b);
		patch(i,b);
	}
}

// small improvement: short cut jump to ret, jmp to jmp
PRIVATE void improv(int ref){
	int i,c1,d1,c2,d2,d;
	for( i=ref+1; i<CP-1; i++ ){
		decode(i,&c1,&d1);
		if( c1==tJmp ){
		    decode(d1,&c2,&d2);
			if( c2 == tJmp )
				patch(i,d2);
			else if( c2==tRet )
				putcode(i,c2,d2);
		}else if( c1==tJt || c1==tJf ){
			d = d1 >> 8;
			decode(i+d,&c2,&d2);
			if( c2 == tJmp )
				patchd(i,d2-i);
		}
	}
}

PRIVATE void tailcall(int ads, int ref){
	int op, a;
	decode(ads,&op,&a);
	if( op == tCall && a == ref )
		putcode(ads,tCallt,a);
	else if( op == tMov ){
		decode(ads-1,&op,&a);
		if( op == tCall && a == ref )
			putcode(ads-1,tCallt,a);
	}
}

// only when function body is not empty
void clean(int fn) {
	int ref, a, op;
	ref = getRef(fn);
	tailcall(CP-2,ref);
	// update fun header to any change of lv
	patcha2(ref,vmax);
    reName(fn);
	improv(ref);
}

// change Callf to Call from CS[a]..CS[b]
void patchCallf(int a, int b){
	int i, c, d;
	for(i=a; i<=b; i++){
		decode(i,&c,&d);
		if( c == tCallf )
			putcode(i,tCall,getRef(d));
	}
}
/*
// relocate DS and ld/st/ads in CS by disp
// assume symtab is loaded, and getExportSym is done
void reLoc(int disp){
	int i, k, a, op, arg;
//	nxsym = getExportSym(exsym);
	for(i=0; i<nxsym; i++){
		k = exsym[i];
		if(getType(k) == tyGVAR){
			a = getArg(k);
			if(a == VSTRING || a == VARRAY)
			   M[getRef(k)] += disp;
		}
	}
	for(i=MAXSYS; i<DS; i++)	// move DS
		M[i+disp] = M[i];
	DS += disp;
	for(i=1; i<CP; i++){		// reloc CS
		decode(i,&op,&arg);
		if(op == tLd || op == tSt || op == tAds)   // ???
			patch(i,arg+disp);
	}
}
*/
